# -*- coding: utf-8 -*-
import sys
import os

project_path = os.path.dirname(
    (os.path.dirname(os.path.realpath(__file__)))
)

# Add the app to path
sys.path.insert(0, project_path)


# -- General configuration ------------------------------------------------
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.coverage',
    'sphinx.ext.viewcode',
]

templates_path = ['_templates']

source_suffix = '.rst'

source_encoding = 'utf-8-sig'

master_doc = 'index'

project = 'Django Lanthanum'
copyright = '2018, Kingston Labs'
author = 'Kingston Labs'

version = '0.1'
release = '0.1.3'

language = 'en'

exclude_patterns = []

pygments_style = 'sphinx'

todo_include_todos = False


# -- Options for HTML output ----------------------------------------------

html_theme = "sphinx_rtd_theme"

# Output file base name for HTML help builder.
htmlhelp_basename = 'DjangoLanthanumDoc'



from decimal import Decimal

from ..schema_fields import (
    ArrayField,
    BooleanField,
    DynamicArrayField,
    Field,
    CharField,
    DecimalField,
    IntegerField,
    TextField
)
from .utils import assert_dict_equal


class TestField:
    def test_schema_basic(self):
        field = Field()
        assert field.schema == {'type': 'string', 'format': 'text'}

    def test_schema(self):
        label = "Test Field"
        default = "Thing"
        required = True
        field = Field(label=label, default=default, required=required)
        assert_dict_equal(
            expected={
                'type': 'string',
                'format': 'text',
                'title': label,
                'default': default
            },
            actual=field.schema
        )

    def test_load_data(self):
        test_data = "A test"
        field = Field()
        loaded_data = field.Meta.python_type(test_data)
        assert loaded_data == test_data


class TestCharField:
    def test_schema_basic(self):
        field = CharField()
        assert field.schema == {'type': 'string', 'format': 'text'}

    def test_schema(self):
        label = "Test Field"
        default = "Thing"
        required = True
        field = CharField(label=label, default=default, required=required)
        assert_dict_equal(
            expected={
                'type': 'string',
                'format': 'text',
                'minLength': 1,
                'title': label,
                'default': default
            },
            actual=field.schema
        )

    def test_schema_min_max_length(self):
        label = "Test Field"
        default = "Thing"
        required = True
        min_length = 3
        max_length = 5
        field = CharField(
            label=label,
            default=default,
            required=required,
            min_length=min_length,
            max_length=max_length
        )
        assert_dict_equal(
            expected={
                'type': 'string',
                'format': 'text',
                'minLength': min_length,
                'maxLength': max_length,
                'title': label,
                'default': default
            },
            actual=field.schema
        )

    def test_schema_choices(self):
        label = "Test Field"
        default = "cat"
        choices = [('cat', 'Cat'), ('dog', 'Dog'), ('fish', 'Fish')]
        required = True
        field = CharField(
            label=label,
            choices=choices,
            default=default,
            required=required
        )
        assert_dict_equal(
            expected={
                'type': 'string',
                'format': 'text',
                'minLength': 1,
                'title': label,
                'default': default,
                'enum': ['cat', 'dog', 'fish'],
                'enumSource': [
                    {
                        'source': [
                            {'value': choice_value, 'title': choice_label}
                            for (choice_value, choice_label) in choices
                        ],
                        'title': '{{item.title}}',
                        'value': '{{item.value}}'
                    }
                ],
            },
            actual=field.schema
        )

    def test_load_data(self):
        test_data = "A test"
        field = CharField()
        loaded_data = field.Meta.python_type(test_data)
        assert loaded_data == test_data


class TestTextField:
    def test_schema_basic(self):
        field = TextField()
        assert_dict_equal(
            expected={'type': 'string', 'format': 'textarea'},
            actual=field.schema
        )

    def test_schema(self):
        label = "Test Field"
        default = "Thing"
        required = True
        field = TextField(label=label, default=default, required=required)
        assert_dict_equal(
            expected={
                'type': 'string',
                'format': 'textarea',
                'minLength': 1,
                'title': label,
                'default': default
            },
            actual=field.schema
        )

    def test_load_data(self):
        test_data = "A test"
        field = TextField()
        loaded_data = field.Meta.python_type(test_data)
        assert loaded_data == test_data


class TestIntegerField:
    def test_schema_basic(self):
        field = IntegerField()
        assert_dict_equal(
            expected={'type': 'integer', 'format': 'number'},
            actual=field.schema
        )

    def test_schema(self):
        label = "Test Field"
        default = 5
        required = True
        field = IntegerField(label=label, default=default, required=required)
        assert_dict_equal(
            expected={
                'type': 'integer',
                'format': 'number',
                'title': label,
                'default': default
            },
            actual=field.schema
        )

    def test_load_data(self):
        test_data = 8
        field = IntegerField()
        loaded_data = field.Meta.python_type(test_data)
        assert loaded_data == test_data


class TestDecimalField:
    def test_schema_basic(self):
        field = DecimalField()
        assert_dict_equal(
            expected={'type': 'decimal', 'format': 'number'},
            actual=field.schema
        )

    def test_schema(self):
        label = "Test Field"
        default = Decimal('5.5')
        required = True
        field = DecimalField(label=label, default=default, required=required)
        assert_dict_equal(
            expected={
                'type': 'decimal',
                'format': 'number',
                'title': label,
                'default': default
            },
            actual=field.schema
        )

    def test_load_data(self):
        test_data = Decimal('3.14')
        field = DecimalField()
        loaded_data = field.Meta.python_type(test_data)
        assert loaded_data == test_data


class TestBooleanField:
    def test_schema_basic(self):
        field = BooleanField()
        assert_dict_equal(
            expected={'type': 'boolean', 'format': 'checkbox'},
            actual=field.schema
        )

    def test_schema(self):
        label = "Test Field"
        default = False
        required = True
        field = BooleanField(label=label, default=default, required=required)
        assert_dict_equal(
            expected={
                'type': 'boolean',
                'format': 'checkbox',
                'title': label,
                'default': default
            },
            actual=field.schema
        )

    def test_load_data(self):
        test_data = True
        field = BooleanField()
        loaded_data = field.Meta.python_type(test_data)
        assert loaded_data == test_data


class TestObjectField:
    def test_schema(self, dog_field, dog_schema):
        assert_dict_equal(
            expected=dog_schema,
            actual=dog_field().schema
        )

    def test_typed_schema(self, dog_field, typed_dog_schema):
        assert_dict_equal(
            expected=typed_dog_schema,
            actual=dog_field().typed_schema
        )

    def test_load_data(self, dog_field, scooby_doo):
        loaded_data = dog_field().Meta.python_type(scooby_doo)
        assert loaded_data.name == scooby_doo['name']
        assert loaded_data.breed == scooby_doo['breed']

    def test_nested_schema(self, person_field, person_schema):
        assert_dict_equal(expected=person_schema, actual=person_field().schema)

    def test_load_nested_data(self, person_field, scooby_doo, shaggy):
        loaded_data = person_field().Meta.python_type(shaggy)
        assert loaded_data.name == shaggy['name']
        assert loaded_data.favourite_dog.name == scooby_doo['name']
        assert loaded_data.favourite_dog.breed == scooby_doo['breed']

    def test_instances_copy_methods(self, dog_field, scooby_doo):
        """
        Methods declared on the field should be available to the instance
        """
        loaded_data = dog_field().Meta.python_type(scooby_doo)
        assert loaded_data.name == scooby_doo['name']
        assert loaded_data.short_name == scooby_doo['name'][:3]


class TestSubClassedObjectField:
    def test_schema(self, parrot_field, parrot_schema):
        assert_dict_equal(
            expected=parrot_schema,
            actual=parrot_field().schema
        )

    def test_typed_schema(self, parrot_field, typed_parrot_schema):
        assert_dict_equal(
            expected=typed_parrot_schema,
            actual=parrot_field().typed_schema
        )

    def test_load_data(self, parrot_field):
        polly = {'name': 'Polly', 'talks': True}
        loaded_data = parrot_field().Meta.python_type(polly)
        assert loaded_data.name == polly['name']
        assert loaded_data.talks == polly['talks']

    def test_instances_copy_parent_methods(self, parrot_field):
        """
        Methods declared on the parent field should be copied to the instance
        """
        polly = {'name': 'Polly', 'talks': True}
        loaded_data = parrot_field().Meta.python_type(polly)
        assert loaded_data.loud_name == polly['name'].upper()


class TestArrayField:
    def test_schema(self, dog_field, dog_schema):
        dog_list_field = ArrayField(base_field=dog_field())
        assert_dict_equal(
            expected={
                'type': 'array',
                'format': 'table',
                'title': 'Dog List',
                'items': dog_schema
            },
            actual=dog_list_field.schema
        )

    def test_load_data(self, dog_field, scooby_doo, snoopy):
        dog_list_field = ArrayField(base_field=dog_field())

        dog_list = [scooby_doo, snoopy]

        loaded_data = dog_list_field.Meta.python_type(dog_list)
        assert len(loaded_data) == 2
        scooby_doo_instance = loaded_data[0]
        snoopy_instance = loaded_data[1]
        assert scooby_doo_instance.name == scooby_doo['name']
        assert scooby_doo_instance.breed == scooby_doo['breed']
        assert snoopy_instance.name == snoopy['name']
        assert snoopy_instance.breed == snoopy['breed']


class TestDynamicArrayField:
    def test_schema(
        self, dog_field, fish_field, typed_dog_schema, typed_fish_schema
    ):
        item_label = "Pet"
        schema_name = "pet_list"
        unique_items = True
        min_items = 2
        max_items = 5

        pet_field = DynamicArrayField(
            schema_name=schema_name,
            item_label=item_label,
            allowed_fields=[dog_field(), fish_field()],
            unique_items=unique_items,
            min_items=min_items,
            max_items=max_items
        )
        assert_dict_equal(
            expected={
                'type': 'array',
                'format': 'tabs',
                'title': schema_name.replace("_", " ").title(),
                'uniqueItems': unique_items,
                'minItems': min_items,
                'maxItems': max_items,
                'items': {
                    'headerTemplate': "{} {{{{i1}}}}.".format(item_label),
                    'oneOf': [typed_dog_schema, typed_fish_schema],
                    'title': item_label
                }
            },
            actual=pet_field.schema
        )

    def test_auto_generated_schema_name(
        self, dog_field, fish_field, typed_dog_schema, typed_fish_schema
    ):
        """
        If no schema name is provided, it should be built from allowed fields
        """
        item_label = "Pet"
        pet_field = DynamicArrayField(
            item_label=item_label,
            allowed_fields=[dog_field(), fish_field()]
        )
        assert pet_field.Meta.schema_name == "one_of_dog_or_fish"
        assert pet_field.schema['title'] == "One Of Dog Or Fish"

    def test_provided_schema_name(
        self, dog_field, fish_field, typed_dog_schema, typed_fish_schema
    ):
        """
        If a schema name is provided, it should be available as class meta
        """
        item_label = "Pet"
        pet_field = DynamicArrayField(
            item_label=item_label,
            schema_name="pet",
            allowed_fields=[dog_field(), fish_field()]
        )
        assert pet_field.Meta.schema_name == "pet"

    def test_load_data(
        self,
        dog_field,
        fish_field,
        typed_dog_schema,
        typed_fish_schema,
        scooby_doo,
        nemo
    ):
        item_label = "Pet"
        pet_field = DynamicArrayField(
            item_label=item_label,
            allowed_fields=[dog_field(), fish_field()]
        )

        pet_list = [
            {
                "schemaName": "dog",
                "data": scooby_doo
            },
            {
                "schemaName": "fish",
                "data": nemo
            }
        ]

        loaded_data = pet_field.Meta.python_type(pet_list)
        assert len(loaded_data) == 2
        scooby_doo_instance = loaded_data[0]
        nemo_instance = loaded_data[1]
        assert scooby_doo_instance.name == scooby_doo['name']
        assert scooby_doo_instance.breed == scooby_doo['breed']
        assert nemo_instance.name == nemo['name']
        assert nemo_instance.salt_water == nemo['salt_water']



import re


def camel_to_underscore(string_in):
    """
    Convert a camel case string to an underscored string
    """
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', string_in)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()


def strip_suffix(string_in, suffix):
    """
    Strip a suffix from a string
    """
    if string_in.endswith(suffix):
        return string_in[:-len(suffix)]
    return string_in


def field_to_schema_name(field_name):
    """
    Convert a field class name to a schema name and remove the Field suffix
    """
    return camel_to_underscore(strip_suffix(field_name, "Field"))



import unittest


def assert_list_equal(expected, actual, ignore_order=True):
    """
    Assert two lists are equal, but also provides the option to ignore order
    """
    assert isinstance(actual, list)

    if ignore_order:
        test_case = unittest.TestCase()
        test_case.assertCountEqual(expected, actual)
    else:
        assert expected == actual


def assert_dict_equal(expected, actual, ignore_list_order=True):
    """
    Iterate over dict and assert each part is equal

    This will cause an assertion error on the first difference, but should be
    more informative than a standard assertion as it will specify exactly
    where the difference is.
    """
    assert isinstance(actual, dict)
    assert expected.keys() == actual.keys()
    for key, value in expected.items():
        if isinstance(value, dict):
            assert_dict_equal(value, actual[key])
        elif isinstance(value, list):
            assert_list_equal(
                value,
                actual[key],
                ignore_order=ignore_list_order
            )
        else:
            assert value == actual[key]



from lanthanum.schema_fields import ObjectField, CharField, DynamicArrayField


class SingleField(ObjectField):
    class Meta:
        schema_name = "single"

    title = CharField(required=True)
    artist = CharField()


class AlbumField(ObjectField):
    title = CharField(required=True)


music_catalog_field = DynamicArrayField(
    allowed_fields=[
        SingleField(),
        AlbumField()
    ],
    item_label="Product"
)



def to_schema_field(field_class):
    """
    Wraps a field class to ensure that it extracts the data from JSON Schema.
    """
    class DynamicJSONField(field_class):
        def prepare_value(self, value):
            """
            Use the raw field data in the JSON field.
            """
            if value is None:
                return value
            return super().prepare_value(getattr(value, '_data', value))

    return DynamicJSONField



from collections import MutableSequence

from .schema_registry import get_python_type


class DynamicObject(object):
    """
    A type for schema objects to subclass
    """
    @classmethod
    def _get_sub_types(cls):
        return {
            name: field_type
            for name, field_type in cls.__dict__.items()
        }

    def __init__(self, data):
        self._data = data
        sub_types = self._get_sub_types()
        for k, v in data.items():
            if k in sub_types:
                setattr(self, k, sub_types[k](v))


class TypedArray(MutableSequence):
    """
    An array with a specified type for each item
    """
    class Meta:
        base_type = str

    def __init__(self, data, **kwargs):
        """
        Initiate list as instance property
        """
        self._data = data
        self._list = list()
        self.extend(list(data))

    def __len__(self):
        return len(self._list)

    def __getitem__(self, i):
        return self._list[i]

    def __delitem__(self, i):
        del self._list[i]

    def __setitem__(self, i, v):
        self._list[i] = self.Meta.base_type(v)

    def insert(self, i, v):
        self._list.insert(i, self.Meta.base_type(v))

    def __str__(self):
        return str(self._list)

    def __repr__(self):
        return str(self._list)


class DynamicArray(MutableSequence):
    """
    An array that includes items of different types
    """
    def __init__(self, data, **kwargs):
        """
        Initiate list as instance property
        """
        self._data = data
        self._list = list()
        self.extend(list(data))

    def __len__(self):
        return len(self._list)

    def __getitem__(self, i):
        return self._list[i]

    def __delitem__(self, i):
        del self._list[i]

    def __setitem__(self, i, v):
        python_type = get_python_type(schema_name=v['schemaName'])
        self._list[i] = python_type(v['data'])

    def insert(self, i, v):
        python_type = get_python_type(schema_name=v['schemaName'])
        self._list.insert(i, python_type(v['data']))

    def __str__(self):
        return str(self._list)

    def __repr__(self):
        return str(self._list)



from django.db import models
from lanthanum.fields import DynamicField

from .schema_fields import music_catalog_field


class RecordShop(models.Model):
    name = models.CharField(max_length=50, unique=True)
    catalog = DynamicField(
        schema_field=music_catalog_field,
        blank=True,
        null=True
    )



import json

from django.core.serializers import serialize
import pytest

from .mock_app.models import RecordShop


@pytest.fixture
def record_catalog():
    return [
        {
            'schemaName': 'single',
            'data': {
                'title': 'Scooby Snacks',
                'artist': 'Scooby Doo'
            }
        },
        {
            'schemaName': 'album',
            'data': {
                'title': 'Who Let The Dogs Out?'
            }
        }
    ]


@pytest.fixture
def alternative_record_catalog():
    return [
        {
            'schemaName': 'single',
            'data': {
                'title': 'D-O-G-G',
                'artist': 'Snoopy'
            }
        },
        {
            'schemaName': 'single',
            'data': {
                'title': 'The Cat in the Hat',
                'artist': 'Top Cat'
            }
        }
    ]


@pytest.fixture
def hmv_instance(record_catalog):
    hmv = RecordShop(name="HMV", catalog=record_catalog)
    hmv.save()
    hmv.refresh_from_db()
    return hmv


@pytest.mark.django_db
class TestFields:
    def test_load_json_to_dynamic_field(self, hmv_instance, record_catalog):
        """
        It should be possible to load json data directly into a dynamic field
        """
        item_1 = hmv_instance.catalog[0]
        assert item_1.title == record_catalog[0]['data']['title']

    def test_empty_data_to_dynamic_field(self, hmv_instance):
        """
        Allow empty dynamic fields
        """
        record_shop = RecordShop(name="OurPrice", catalog=None)
        record_shop.save()
        record_shop.refresh_from_db()
        assert record_shop.catalog is None

    def test_serialize_dynamic_field(self, hmv_instance, record_catalog):
        """
        Serialize the dynamic field to json
        """
        json_record_shops = serialize(
            format='json', queryset=RecordShop.objects.all()
        )
        assert json.loads(json_record_shops)[0]['fields'] == {
            "name": "HMV",
            "catalog": record_catalog
        }

    def test_add_form_field(
        self, alternative_record_catalog, record_shop_form_class
    ):
        record_shop_form = record_shop_form_class(
            {'name': 'virgin', 'catalog': alternative_record_catalog}
        )

        assert record_shop_form.is_valid()

        record_shop_form.save()
        virgin_instance = RecordShop.objects.get(name="virgin")
        assert len(virgin_instance.catalog) == len(alternative_record_catalog)

        item_1 = virgin_instance.catalog[0]
        assert item_1.title == alternative_record_catalog[0]['data']['title']
        assert item_1.artist == alternative_record_catalog[0]['data']['artist']

        item_2 = virgin_instance.catalog[1]
        assert item_2.title == alternative_record_catalog[1]['data']['title']
        assert item_2.artist == alternative_record_catalog[1]['data']['artist']

    def test_update_form_field(
        self, hmv_instance, alternative_record_catalog, record_shop_form_class
    ):
        record_shop_form = record_shop_form_class(
            {
                'name': 'HMV',
                'catalog': alternative_record_catalog
            },
            instance=hmv_instance
        )

        assert record_shop_form.is_valid()

        record_shop_form.save()

        hmv_instance.refresh_from_db()
        assert len(hmv_instance.catalog) == len(alternative_record_catalog)

        item_1 = hmv_instance.catalog[0]
        assert item_1.title == alternative_record_catalog[0]['data']['title']
        assert item_1.artist == alternative_record_catalog[0]['data']['artist']

        item_2 = hmv_instance.catalog[1]
        assert item_2.title == alternative_record_catalog[1]['data']['title']
        assert item_2.artist == alternative_record_catalog[1]['data']['artist']

    def test_render_empty_form_field_widget(self, record_shop_form_class):
        record_shop_form = record_shop_form_class()

        widget = record_shop_form.fields['catalog'].widget
        assert widget.template_name == "lanthanum/_json_editor_widget.html"
        # Just check the widget is able to render something
        assert widget.render(name="catalog", value=None)



import os
from setuptools import find_packages, setup

with open(os.path.join(os.path.dirname(__file__), 'README.rst')) as readme:
    README = readme.read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))


tests_require = [
    'pytest>=4.0.1',
    'pytest-cov>=2.6.0',
    'pytest-django>=3.4.4'
]

install_requires = [
    'django>=2.0',
    'django-admin-json-editor>=0.1.5',
    'jsonschema>=2.6.0',
    'psycopg2==2.7.5'
]


setup(
    name='django-lanthanum',
    version='0.1.3',
    description='Dynamic JSON Schema Fields for Django Postgres Models',
    long_description=README,
    author='Kingston Labs',
    author_email='tom@kingstonlabs.co.uk',
    license='BSD',
    url='https://github.com/kingstonlabs/lanthanum',
    packages=find_packages(),
    include_package_data=True,
    classifiers=[
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Framework :: Django',
        'Framework :: Django :: 2.0',
        'Framework :: Django :: 2.1',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    ],
    install_requires=install_requires,
    tests_require=tests_require,
)



from django.forms import ModelForm
import pytest

from ..schema_fields import BooleanField, CharField, ObjectField
from .mock_app.models import RecordShop


@pytest.fixture
def scooby_doo():
    return {'name': 'Scooby Doo', 'breed': 'Daschund'}


@pytest.fixture
def snoopy():
    return {'name': 'Snoopy', 'breed': 'Beagle'}


@pytest.fixture
def shaggy(scooby_doo):
    return {'name': 'Shaggy', 'favourite_dog': scooby_doo}


@pytest.fixture
def nemo():
    return {'name': 'Nemo', 'salt_water': True}


@pytest.fixture
def dog_field():
    class DogField(ObjectField):
        name = CharField(required=True)
        breed = CharField()

        @property
        def short_name(self):
            return self.name[:3]

    return DogField


@pytest.fixture
def dog_schema():
    return {
        'type': 'object',
        'title': 'Dog',
        'properties': {
            'name': {
                'title': 'Name',
                'type': 'string',
                'format': 'text',
                'minLength': 1
            },
            'breed': {
                'title': 'Breed',
                'type': 'string',
                'format': 'text'
            }
        },
        'required': ['name']
    }


@pytest.fixture
def typed_dog_schema(dog_schema):
    schema_name = 'dog'
    return {
        'type': 'object',
        'title': 'Dog',
        'properties': {
            'schemaName': {
                'title': 'Schema Name',
                'const': schema_name,
                'type': 'string',
                'default': schema_name,
                'template': schema_name
            },
            'data': dog_schema
        },
        "defaultProperties": ["data", "schemaName"],
        "required": ['data', 'schemaName']
    }


@pytest.fixture
def person_field(dog_field):
    class PersonField(ObjectField):
        name = CharField(required=True)
        favourite_dog = dog_field()
    return PersonField


@pytest.fixture
def person_schema(dog_schema):
    dog_schema['title'] = 'Favourite Dog'
    return {
        'type': 'object',
        'title': 'Person',
        'properties': {
            'name': {
                'title': 'Name',
                'type': 'string',
                'format': 'text',
                'minLength': 1
            },
            'favourite_dog': dog_schema
        },
        'required': ['name']
    }


@pytest.fixture
def fish_field():
    class FishField(ObjectField):
        name = CharField(required=True)
        salt_water = BooleanField(default=False)
    return FishField


@pytest.fixture
def fish_schema():
    return {
        'type': 'object',
        'title': 'Fish',
        'properties': {
            'name': {
                'title': 'Name',
                'type': 'string',
                'format': 'text',
                'minLength': 1
            },
            'salt_water': {
                'title': 'Salt Water',
                'type': 'boolean',
                'format': 'checkbox',
                'default': False
            }
        },
        'required': ['name']
    }


@pytest.fixture
def typed_fish_schema(fish_schema):
    schema_name = 'fish'
    return {
        'type': 'object',
        'title': 'Fish',
        'properties': {
            'schemaName': {
                'title': 'Schema Name',
                'const': schema_name,
                'type': 'string',
                'default': schema_name,
                'template': schema_name
            },
            'data': fish_schema
        },
        "defaultProperties": ["data", "schemaName"],
        "required": ['data', 'schemaName']
    }


@pytest.fixture
def generic_animal_field():

    class GenericAnimalField(ObjectField):
        name = CharField(required=True)

        class Meta:
            abstract = True

        @property
        def loud_name(self):
            return self.name.upper()

    return GenericAnimalField


@pytest.fixture
def parrot_field(generic_animal_field):
    """
    A parrot field subclasses generic animal field
    """

    class ParrotField(generic_animal_field):
        talks = BooleanField(required=True)

        class Meta:
            schema_name = "squawker"

    return ParrotField


@pytest.fixture
def parrot_schema():
    return {
        'type': 'object',
        'title': 'Squawker',
        'properties': {
            'name': {
                'title': 'Name',
                'type': 'string',
                'format': 'text',
                'minLength': 1
            },
            'talks': {
                'title': 'Talks',
                'type': 'boolean',
                'format': 'checkbox'
            }
        },
        'required': ['name', 'talks']
    }


@pytest.fixture
def typed_parrot_schema(parrot_schema):
    schema_name = 'squawker'
    return {
        'type': 'object',
        'title': 'Squawker',
        'properties': {
            'schemaName': {
                'title': 'Schema Name',
                'const': schema_name,
                'type': 'string',
                'default': schema_name,
                'template': schema_name
            },
            'data': parrot_schema
        },
        "defaultProperties": ["data", "schemaName"],
        "required": ['data', 'schemaName']
    }


@pytest.fixture
def record_shop_form_class():
    class RecordShopForm(ModelForm):
        class Meta:
            model = RecordShop
            fields = ['name', 'catalog']

    return RecordShopForm



from django.contrib.postgres.fields import JSONField
from django.contrib.postgres.forms import JSONField as JSONFormField

from .form_fields import to_schema_field
from .widgets import JSONEditorWidget


class DynamicField(JSONField):
    description = "A dynamic field for json schema data"

    def __init__(self, *args, **kwargs):
        self.schema_field = kwargs.pop("schema_field")
        self.output_type = self.schema_field.Meta.python_type
        super().__init__(*args, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        if self.schema_field is not None:
            kwargs['schema_field'] = self.schema_field
        return name, path, args, kwargs

    def from_db_value(self, value, expression, connection):
        """
        Convert the data coming out of the database into the correct type.
        """
        if value is None:
            return value

        parsed_value = self.output_type(value)

        return parsed_value

    def value_to_string(self, obj):
        """
        Convert object to data for data dumps.
        """
        value = self.value_from_object(obj)
        return value._data

    def formfield(self, **kwargs):
        """
        Provide a JSON Editor Widget to work with the data.
        """
        widget = JSONEditorWidget(
            self.schema_field.schema,
            collapsed=False
        )
        defaults = {'form_class': JSONFormField, 'widget': widget}
        defaults.update(kwargs)
        defaults['form_class'] = to_schema_field(defaults['form_class'])
        return super().formfield(**defaults)



import copy
import json

from django_admin_json_editor import JSONEditorWidget
from django.utils.safestring import mark_safe
from django.template.loader import render_to_string


class JSONEditorWidget(JSONEditorWidget):
    template_name = 'lanthanum/_json_editor_widget.html'

    def render(self, name, value, attrs=None, renderer=None):
        """
        Fix the JSON Editor widget by doing a standard json dump for dict data

        This will not convert booleans to ints like the standard JSON Editor.
        """
        if callable(self._schema):
            schema = self._schema(self)
        else:
            schema = copy.copy(self._schema)

        schema['title'] = ' '
        schema['options'] = {'collapsed': int(self._collapsed)}

        context = {
            'name': name,
            'schema': json.dumps(schema),
            'data': value,
            'sceditor': int(self._sceditor),
        }
        return mark_safe(render_to_string(self.template_name, context))



from decimal import Decimal
import logging

from .field_types import DynamicArray, DynamicObject, TypedArray
from .schema_registry import schema_registry
from .utils import field_to_schema_name, strip_suffix


logger = logging.getLogger(__name__)


class Field(object):
    """
    Basic JSON schema field definition
    """
    class Meta:
        python_type = str
        schema_type = 'string'
        schema_format = 'text'
        schema_name = 'field'
        abstract = True

        @classmethod
        def get_default_label(cls):
            return None

    def __new__(cls, *args, **kwargs):
        """
        Configure the new field
        """
        new_class = super().__new__(cls)

        # Inherit meta attributes from base classes
        merged_meta_dict = {}
        for base in reversed(cls.mro()):
            if hasattr(base, 'Meta'):
                merged_meta_dict.update(base.Meta.__dict__)
        # Abstract should always default to False unless explicitly specified
        # Schema name should not be inherited
        merged_meta_dict.update({
            'abstract': getattr(new_class.Meta, 'abstract', False),
            'schema_name': (
                getattr(new_class.Meta, 'schema_name', None) or
                field_to_schema_name(new_class.__class__.__name__)
            )
        })

        new_class.Meta = type('Meta', (), merged_meta_dict)

        key = new_class.Meta.schema_name

        is_abstract = getattr(cls.Meta, 'abstract', False)
        if not is_abstract and key not in schema_registry:
            schema_registry[key] = new_class
        return new_class

    def __init__(self, **kwargs):
        """
        Configure options: label, default and required
        """
        self._label = kwargs.get('label', self.Meta.get_default_label())
        self._default = kwargs.get('default')
        self._required = kwargs.get('required', False)

    @property
    def schema(self):
        """
        The JSON Schema for basic fields
        """
        schema = {}
        if self.Meta.schema_type is not None:
            schema['type'] = self.Meta.schema_type
        if self.Meta.schema_format is not None:
            schema['format'] = self.Meta.schema_format
        if self._label is not None:
            schema['title'] = self._label
        if self._default is not None:
            schema['default'] = self._default
        return schema

    @property
    def typed_schema(self):
        """
        Add data typing to the schema by wrapping it with metadata

        This is useful for validating OneOf Schemas because it includes the
        schema type as a constant.
        """
        schema = {
            'type': 'object',
            'properties': {
                'schemaName': {
                    'title': 'Schema Name',
                    'const': self.Meta.schema_name,
                    'type': 'string',
                    'default': self.Meta.schema_name,
                    # JSON Editor isn't very good at making a const field, so
                    # set to a static template
                    'template': self.Meta.schema_name
                },
                'data': self.schema
            },
            "defaultProperties": ["data", "schemaName"],
            "required": ['data', 'schemaName']
        }
        if self._label is not None:
            schema['title'] = self._label
        return schema


class CharField(Field):
    """
    JSON Schema Field for a simple string input
    """
    class Meta:
        python_type = str
        schema_type = 'string'
        schema_format = 'text'
        schema_name = 'charfield'
        abstract = False

    def __init__(self, **kwargs):
        """
        Configure with choices, min_length and max_length options
        """
        super().__init__(**kwargs)
        self._choices = kwargs.pop("choices", None)
        self._min_length = kwargs.pop("min_length", None)
        self._max_length = kwargs.pop("max_length", None)

    @property
    def schema(self):
        """
        CharField schema including choices and min / max length
        """
        schema = super().schema
        if self._choices:
            schema['enumSource'] = [
                {
                    'source': [
                        {'value': value, 'title': label}
                        for (value, label) in self._choices
                    ],
                    'title': '{{item.title}}',
                    'value': '{{item.value}}'
                }
            ]
            schema['enum'] = [value for (value, label) in self._choices]
        if self._required and self._min_length is None:
            schema['minLength'] = 1
        if self._min_length:
            schema['minLength'] = self._min_length
        if self._max_length:
            schema['maxLength'] = self._max_length
        return schema


class TextField(Field):
    """
    JSON Schema field for larger text inputs
    """
    class Meta:
        python_type = str
        schema_type = 'string'
        schema_format = 'textarea'
        schema_name = 'textfield'
        abstract = False

    @property
    def schema(self):
        """
        Just set min length for required fields
        """
        schema = super().schema
        if self._required:
            schema['minLength'] = 1
        return schema


class BooleanField(Field):
    """
    JSON Schema for boolean inputs
    """
    class Meta:
        python_type = bool
        schema_type = 'boolean'
        schema_format = 'checkbox'
        schema_name = 'booleanfield'
        abstract = False


class IntegerField(Field):
    """
    JSON Schema for integer inputs
    """
    class Meta:
        python_type = int
        schema_type = 'integer'
        schema_format = 'number'
        schema_name = 'integerfield'
        abstract = False


class DecimalField(Field):
    """
    JSON Schema for decimal inputs
    """
    class Meta:
        python_type = Decimal
        schema_type = 'decimal'
        schema_format = 'number'
        schema_name = 'numberfield'
        abstract = False


class ObjectField(Field):
    """
    JSON Schema for object inputs
    """
    class Meta:
        python_type = DynamicObject
        schema_type = 'object'
        schema_format = None
        schema_name = None
        abstract = True

        @classmethod
        def get_default_label(cls):
            return cls.schema_name.replace("_", " ").title()

    def __new__(cls, *args, **kwargs):
        """
        Configure the new field with the sub fields specified

        This dynamically generates a new python type to fit the JSON Schema.
        """
        schema_name = (
            cls.Meta.schema_name or
            field_to_schema_name(cls.__name__)
        )

        if schema_name in schema_registry:
            return schema_registry[schema_name]

        new_class = super().__new__(cls, *args, **kwargs)

        sub_fields = {}
        python_type_dict = {}
        for base in reversed(cls.mro()):
            sub_fields.update({
                name: field
                for name, field in base.__dict__.items()
                if isinstance(field, Field)
            })
            python_type_dict.update({
                name: prop
                for name, prop in base.__dict__.items()
                if not name.startswith("__")
            })

        new_class._sub_fields = sub_fields
        new_class._required_field_names = [
            name for name, field in new_class._sub_fields.items()
            if field._required
        ]

        python_type_dict.update({
            key: field.Meta.python_type
            for key, field in new_class._sub_fields.items()
        })
        # Make sure the schema name is available to the resulting type
        python_type_dict['schema_name'] = schema_name

        python_type = type(
            "{}Type".format(strip_suffix(cls.__name__, "Field")),
            (DynamicObject,),
            python_type_dict
        )

        # Inherit meta attributes from base classes
        merged_meta_dict = {}
        for base in reversed(cls.mro()):
            if hasattr(base, 'Meta'):
                merged_meta_dict.update(base.Meta.__dict__)
        merged_meta_dict.update({
            'python_type': python_type,
            'schema_name': schema_name,
            'schema_type': 'object',
            'schema_format': None,
            'abstract': False
        })

        new_class.Meta = type('Meta', (), merged_meta_dict)

        schema_registry[schema_name] = new_class
        return new_class

    @property
    def schema(self):
        """
        Build the schema by iterating over each of the sub fields.
        """
        schema = super().schema
        schema['properties'] = {}
        for name, sub_field in self._sub_fields.items():
            sub_field_schema = sub_field.schema.copy()
            sub_field_schema['title'] = name.title().replace("_", " ")
            schema['properties'][name] = sub_field_schema
        schema['required'] = self._required_field_names
        return schema


class ArrayField(Field):
    """
    JSON Schema for simple array inputs
    """
    class Meta:
        python_type = TypedArray
        schema_type = 'array'
        schema_format = 'table'
        schema_name = None
        abstract = True

        @classmethod
        def get_default_label(cls):
            simple_label = strip_suffix(cls.schema_name, "_array")
            return "{} List".format(simple_label.replace("_", " ").title())

    def __new__(cls, *args, **kwargs):
        """
        Configure the new field with the base field specified

        This dynamically generates a python array type to contain each item.
        """
        base_field = kwargs.pop('base_field', Field)

        schema_name = cls.Meta.schema_name or "{}_array".format(
            base_field.Meta.schema_name
        )
        if schema_name in schema_registry:
            return schema_registry[schema_name]

        new_class = super().__new__(cls, *args, **kwargs)
        new_class._base_field = base_field

        array_type_meta = type(
            'Meta',
            (),
            {'base_type': base_field.Meta.python_type}
        )
        python_type = type(
            "{}ArrayType".format(
                strip_suffix(base_field.__class__.__name__, "Field")
            ),
            (TypedArray,),
            {'Meta': array_type_meta}
        )

        # Inherit meta attributes from base classes
        merged_meta_dict = {}
        for base in reversed(cls.mro()):
            if hasattr(base, 'Meta'):
                merged_meta_dict.update(base.Meta.__dict__)
        merged_meta_dict.update({
            'python_type': python_type,
            'schema_name': schema_name,
            'schema_type': 'array',
            'schema_format': 'table',
            'abstract': False
        })
        new_class.Meta = type('Meta', (), merged_meta_dict)

        schema_registry[schema_name] = new_class
        return new_class

    @property
    def schema(self):
        """
        Build the schema by iterating over each of the sub fields.
        """
        schema = super().schema
        schema['items'] = self._base_field.schema
        return schema


class DynamicArrayField(Field):
    """
    An array of items that may be of different types

    This stores the typed schema for each item so that it can be processed
    properly when converted into python.
    """
    class Meta:
        python_type = DynamicArray
        schema_type = 'array'
        schema_format = 'tabs'
        schema_name = None
        abstract = False

        @classmethod
        def get_default_label(cls):
            return cls.schema_name.replace("_", " ").title()

    def __new__(cls, *args, **kwargs):
        """
        Configure the new field with the allowed fields specified

        This generates a dynamic python array type to contain each item.
        """
        allowed_fields = kwargs.pop('allowed_fields')
        schema_name = (
            kwargs.pop('schema_name', None) or
            "one_of_{}".format("_or_".join([
                field.Meta.schema_name
                for field in allowed_fields
            ]))
        )

        if schema_name in schema_registry:
            return schema_registry[schema_name]

        new_class = super().__new__(cls, *args, **kwargs)
        new_class._allowed_fields = allowed_fields

        # Inherit meta attributes from base classes
        merged_meta_dict = {}
        for base in reversed(cls.mro()):
            if hasattr(base, 'Meta'):
                merged_meta_dict.update(base.Meta.__dict__)
        merged_meta_dict.update({
                'python_type': cls.Meta.python_type,
                'schema_name': schema_name,
                'schema_type': 'array',
                'schema_format': 'tabs',
                'abstract': False
            })
        new_class.Meta = type('Meta', (), merged_meta_dict)

        schema_registry[schema_name] = new_class
        return new_class

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._unique_items = kwargs.get("unique_items")
        self._max_items = kwargs.get("max_items")
        self._min_items = kwargs.get("min_items")
        self._item_label = kwargs.get("item_label", "Item")

    @property
    def schema(self):
        schema = super().schema

        schema['items'] = {
            'title': self._item_label,
            'headerTemplate': "{} {{{{i1}}}}.".format(self._item_label),
            'oneOf': [field.typed_schema for field in self._allowed_fields]
        }

        if self._unique_items is not None:
            schema['uniqueItems'] = self._unique_items
        if self._min_items is not None:
            schema['minItems'] = self._min_items
        if self._max_items is not None:
            schema['maxItems'] = self._max_items
        return schema



import pytest

from ..field_types import DynamicObject, TypedArray, DynamicArray


@pytest.fixture
def dog_type():
    class DogType(DynamicObject):
        name = str
        breed = str
    return DogType


@pytest.fixture
def fish_type():
    class FishType(DynamicObject):
        name = str
        salt_water = bool
    return FishType


@pytest.fixture
def person_type(dog_type):
    class PersonType(DynamicObject):
        name = str
        favourite_dog = dog_type
    return PersonType


@pytest.fixture
def dog_array_type(dog_type):
    class DogArrayType(TypedArray):
        class Meta:
            base_type = dog_type
    return DogArrayType


class TestDynamicObject:
    def test_simple_type(self, dog_type, scooby_doo):
        dog_instance = dog_type(scooby_doo)
        assert dog_instance.name == scooby_doo['name']
        assert dog_instance.breed == scooby_doo['breed']

    def test_nested_type(self, person_type, shaggy, scooby_doo):
        person_instance = person_type(shaggy)
        assert person_instance.name == shaggy['name']
        assert person_instance.favourite_dog.name == scooby_doo['name']
        assert person_instance.favourite_dog.breed == scooby_doo['breed']


class TestTypedArray:
    def test_simple_array(self, dog_array_type, scooby_doo, snoopy):
        dog_array = dog_array_type([scooby_doo, snoopy])

        assert len(dog_array) == 2
        scooby_doo_instance = dog_array[0]
        snoopy_instance = dog_array[1]
        assert scooby_doo_instance.name == scooby_doo['name']
        assert scooby_doo_instance.breed == scooby_doo['breed']
        assert snoopy_instance.name == snoopy['name']
        assert snoopy_instance.breed == snoopy['breed']
        assert str(dog_array) == str([scooby_doo_instance, snoopy_instance])
        assert repr(dog_array) == str([scooby_doo_instance, snoopy_instance])

    def test_simple_array_update(self, dog_array_type, scooby_doo, snoopy):
        dog_array = dog_array_type([scooby_doo, snoopy])

        assert len(dog_array) == 2
        dog_array[1] = scooby_doo
        assert len(dog_array) == 2

        item_2 = dog_array[1]
        assert item_2.name == scooby_doo['name']
        assert item_2.breed == scooby_doo['breed']

    def test_simple_array_del(self, dog_array_type, scooby_doo, snoopy):
        dog_array = dog_array_type([scooby_doo, snoopy])

        assert len(dog_array) == 2
        del dog_array[1]
        assert len(dog_array) == 1


class TestDynamicArray:
    @pytest.fixture(autouse=True)
    def schema_registry(self, monkeypatch, dog_type, fish_type):
        class MockDogSchema(object):
            class Meta:
                python_type = dog_type

        class MockFishSchema(object):
            class Meta:
                python_type = fish_type

        monkeypatch.setattr(
            'lanthanum.schema_registry.schema_registry',
            {'dog': MockDogSchema, 'fish': MockFishSchema}
        )

    def test_simple_array(self, scooby_doo, nemo):
        pet_array = DynamicArray(
            [
                {'schemaName': 'dog', 'data': scooby_doo},
                {'schemaName': 'fish', 'data': nemo}
            ]
        )

        assert len(pet_array) == 2
        scooby_doo_instance = pet_array[0]
        nemo_instance = pet_array[1]
        assert scooby_doo_instance.name == scooby_doo['name']
        assert scooby_doo_instance.breed == scooby_doo['breed']
        assert nemo_instance.name == nemo['name']
        assert nemo_instance.salt_water == nemo['salt_water']
        assert str(pet_array) == str([scooby_doo_instance, nemo_instance])
        assert repr(pet_array) == str([scooby_doo_instance, nemo_instance])

    def test_simple_array_update(self, scooby_doo, snoopy, nemo):
        pet_array = DynamicArray(
            [
                {'schemaName': 'dog', 'data': scooby_doo},
                {'schemaName': 'fish', 'data': nemo}
            ]
        )

        assert len(pet_array) == 2
        pet_array[1] = {'schemaName': 'dog', 'data': snoopy}
        item_2 = pet_array[1]
        assert item_2.name == snoopy['name']
        assert item_2.breed == snoopy['breed']

    def test_simple_array_del(self, scooby_doo, nemo):
        pet_array = DynamicArray(
            [
                {'schemaName': 'dog', 'data': scooby_doo},
                {'schemaName': 'fish', 'data': nemo}
            ]
        )

        assert len(pet_array) == 2
        del pet_array[1]
        assert len(pet_array) == 1
